@@ -1,7 +1,7 @@
 {-# LANGUAGE MagicHash, RankNTypes, PolyKinds, GADTs, DataKinds,
              FlexibleContexts, FlexibleInstances,
              TypeFamilies, TypeOperators, TypeFamilyDependencies,
-             UndecidableInstances, TypeInType, ConstraintKinds,
+             UndecidableInstances, ConstraintKinds,
              ScopedTypeVariables, TypeApplications, AllowAmbiguousTypes,
              PatternSynonyms, ViewPatterns #-}
 
@@ -10,7 +10,7 @@
 -- Module      :  Data.Singletons.Internal
 -- Copyright   :  (C) 2013 Richard Eisenberg
 -- License     :  BSD-style (see LICENSE)
--- Maintainer  :  Richard Eisenberg (rae@cs.brynmawr.edu)
+-- Maintainer  :  Ryan Scott
 -- Stability   :  experimental
 -- Portability :  non-portable
 --
@@ -26,7 +26,7 @@ module Data.Singletons.Internal (
   , Proxy(..)
   ) where
 
-import Data.Kind
+import Data.Kind (Type)
 import Unsafe.Coerce
 import Data.Proxy ( Proxy(..) )
 import GHC.Exts ( Proxy#, Constraint )
@@ -46,12 +46,12 @@ type SameKind (a :: k) (b :: k) = (() :: Constraint)
 ----------------------------------------------------------------------
 
 -- | The singleton kind-indexed data family.
-data family Sing (a :: k)
+data family Sing :: k -> Type
 
 -- | A 'SingI' constraint is essentially an implicitly-passed singleton.
 -- If you need to satisfy this constraint with an explicit singleton, please
 -- see 'withSingI' or the 'Sing' pattern synonym.
-class SingI (a :: k) where
+class SingI a where
   -- | Produce the singleton explicitly. You will likely need the @ScopedTypeVariables@
   -- extension to use this method the way you want.
   sing :: Sing a
@@ -63,7 +63,8 @@ class SingI (a :: k) where
 --
 -- As a __pattern__: Matches on an explicit @Sing a@ witness bringing
 -- an implicit @SingI a@ constraint into scope.
-pattern Sing :: forall (a :: k). () => SingI a => Sing a
+{-# COMPLETE Sing #-}
+pattern Sing :: forall k (a :: k). () => SingI a => Sing a
 pattern Sing <- (singInstance -> SingInstance)
   where Sing = sing
 
@@ -88,7 +89,7 @@ class SingKind k where
   -- | Get a base type from the promoted kind. For example,
   -- @Demote Bool@ will be the type @Bool@. Rarely, the type and kind do not
   -- match. For example, @Demote Nat@ is @Natural@.
-  type Demote k = (r :: *) | r -> k
+  type Demote k = (r :: Type) | r -> k
 
   -- | Convert a singleton to its unrefined version.
   fromSing :: Sing (a :: k) -> Demote k
@@ -137,6 +138,7 @@ data SomeSing k where
 --     'withSomeSing' bool2 $ \singBool2 ->
 --       'SomeSing' (singBool1 %&& singBool2)
 -- @
+{-# COMPLETE FromSing #-}
 pattern FromSing :: SingKind k => forall (a :: k). Sing a -> Demote k
 pattern FromSing sng <- ((\demotedVal -> withSomeSing demotedVal SomeSing) -> SomeSing sng)
   where FromSing sng = fromSing sng
@@ -153,7 +155,7 @@ data SingInstance (a :: k) where
 newtype DI a = Don'tInstantiate (SingI a => SingInstance a)
 
 -- | Get an implicit singleton (a 'SingI' instance) from an explicit one.
-singInstance :: forall (a :: k). Sing a -> SingInstance a
+singInstance :: forall k (a :: k). Sing a -> SingInstance a
 singInstance s = with_sing_i SingInstance
   where
     with_sing_i :: (SingI a => SingInstance a) -> SingInstance a
@@ -167,11 +169,11 @@ singInstance s = with_sing_i SingInstance
 -- between term-level arrows and this type-level arrow is that at the term
 -- level applications can be unsaturated, whereas at the type level all
 -- applications have to be fully saturated.
-data TyFun :: * -> * -> *
+data TyFun :: Type -> Type -> Type
 
 -- | Something of kind `a ~> b` is a defunctionalized type function that is
 -- not necessarily generative or injective.
-type a ~> b = TyFun a b -> *
+type a ~> b = TyFun a b -> Type
 infixr 0 ~>
 
 -- | Type level function application
@@ -239,7 +241,7 @@ newtype instance Sing (f :: k1 ~> k2) =
   SLambda { applySing :: forall t. Sing t -> Sing (f @@ t) }
 
 -- | An infix synonym for `applySing`
-(@@) :: forall (f :: k1 ~> k2) (t :: k1). Sing f -> Sing t -> Sing (f @@ t)
+(@@) :: forall k1 k2 (f :: k1 ~> k2) (t :: k1). Sing f -> Sing t -> Sing (f @@ t)
 (@@) = applySing
 
 -- | Note that this instance's 'toSing' implementation crucially relies on the fact
@@ -401,7 +403,7 @@ withSing f = f sing
 -- property.  If the singleton does not satisfy the property, then the function
 -- returns 'Nothing'. The property is expressed in terms of the underlying
 -- representation of the singleton.
-singThat :: forall (a :: k). (SingKind k, SingI a)
+singThat :: forall k (a :: k). (SingKind k, SingI a)
          => (Demote k -> Bool) -> Maybe (Sing a)
 singThat p = withSing $ \x -> if p (fromSing x) then Just x else Nothing
 
@@ -425,4 +427,4 @@ singByProxy# _ = sing
 -- >>> demote @(Nothing :: Maybe Ordering)
 -- Nothing
 demote :: forall a. (SingKind (KindOf a), SingI a) => Demote (KindOf a)
-demote = fromSing (sing @(KindOf a) @a)
+demote = fromSing (sing @a)
