@@ -1,14 +1,18 @@
-{-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE DataKinds #-}
+{-# LANGUAGE GADTs #-}
+{-# LANGUAGE InstanceSigs #-}
+{-# LANGUAGE PolyKinds #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TypeInType #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.Singletons.Prelude.IsString
 -- Copyright   :  (C) 2017 Ryan Scott
 -- License     :  BSD-style (see LICENSE)
--- Maintainer  :  Richard Eisenberg (rae@cs.brynmawr.edu)
+-- Maintainer  :  Ryan Scott
 -- Stability   :  experimental
 -- Portability :  non-portable
 --
@@ -23,6 +27,10 @@ module Data.Singletons.Prelude.IsString (
   FromStringSym0, FromStringSym1
   ) where
 
+import Data.Functor.Const
+import Data.Functor.Identity
+import Data.Singletons.Prelude.Const
+import Data.Singletons.Prelude.Identity
 import Data.Singletons.Single
 import Data.Singletons.TypeLits ()   -- for the IsString instance!
 import GHC.TypeLits (Symbol)
@@ -32,6 +40,14 @@ $(singletonsOnly [d|
   --    extension (-XOverloadedStrings in GHC).
   class IsString a where
       fromString :: Symbol -> a
+
+  -- deriving instance IsString a => IsString (Const a (b :: k))
+  instance IsString a => IsString (Const a (b :: k)) where
+    fromString x = Const (fromString x)
+
+  -- deriving instance IsString a => IsString (Identity a)
+  instance IsString a => IsString (Identity a) where
+    fromString x = Identity (fromString x)
   |])
 
 -- PIsString instance
