@@ -1,12 +1,14 @@
-{-# LANGUAGE TemplateHaskell, ScopedTypeVariables, TypeInType, TypeOperators,
-             TypeFamilies, GADTs, UndecidableInstances #-}
+{-# LANGUAGE TemplateHaskell, ScopedTypeVariables, TypeOperators,
+             TypeFamilies, GADTs, UndecidableInstances, InstanceSigs,
+             DataKinds, PolyKinds #-}
+{-# OPTIONS_GHC -Wno-orphans #-}
 
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.Singletons.Prelude.List.NonEmpty
 -- Copyright   :  (C) 2016 Richard Eisenberg
 -- License     :  BSD-style (see LICENSE)
--- Maintainer  :  Richard Eisenberg (rae@cs.brynmawr.edu)
+-- Maintainer  :  Ryan Scott
 -- Stability   :  experimental
 -- Portability :  non-portable
 --
@@ -145,14 +147,16 @@ module Data.Singletons.Prelude.List.NonEmpty (
   XorSym0, XorSym1
   ) where
 
+import Control.Monad.Zip
 import Data.List.NonEmpty
-import Data.Singletons.Prelude.List.NonEmpty.Internal
+import Data.Singletons.Prelude.List.Internal.Disambiguation
 import Data.Singletons.Prelude.Instances
 import Data.Singletons.Prelude.Base hiding ( MapSym0, MapSym1, MapSym2, Map, sMap )
 import Data.Singletons.Prelude.Maybe
 import Data.Singletons.Prelude.Num
 import Data.Singletons.Prelude.Bool
 import Data.Singletons.Prelude.Eq
+import Data.Singletons.Prelude.Monad.Zip
 import Data.Singletons.Prelude.Ord
 import Data.Singletons.Prelude.Function
 import Data.Function
@@ -172,13 +176,12 @@ $(singletonsOnly [d|
   instance MonadFix NonEmpty where
     mfix f = case fix (f . head) of
                ~(x :| _) -> x :| mfix (tail . f)
+  -}
 
-  -- | @since 4.9.0.0
   instance MonadZip NonEmpty where
     mzip     = zip
     mzipWith = zipWith
     munzip   = unzip
-  -}
 
   -- needed to implement other functions
   fmap :: (a -> b) -> NonEmpty a -> NonEmpty b
