@@ -23,16 +23,21 @@ import Data.Singletons.Deriving.Ord
 import Data.Singletons.Deriving.Bounded
 import Data.Singletons.Deriving.Enum
 import Data.Singletons.Deriving.Show
+import Data.Singletons.Deriving.Util
 import Data.Singletons.Partition
 import Data.Singletons.Util
 import Data.Singletons.Syntax
 import Prelude hiding (exp)
 import Control.Applicative (Alternative(..))
+import Control.Arrow (second)
 import Control.Monad
 import Control.Monad.Trans.Class (MonadTrans(..))
 import Control.Monad.Trans.Maybe
+import Control.Monad.Writer
 import qualified Data.Map.Strict as Map
 import Data.Map.Strict ( Map )
+import qualified Data.Set as Set
+import Data.Set ( Set )
 import Data.Maybe
 import qualified GHC.LanguageExtensions.Type as LangExt
 
@@ -41,7 +46,7 @@ import qualified GHC.LanguageExtensions.Type as LangExt
 genPromotions :: DsMonad q => [Name] -> q [Dec]
 genPromotions names = do
   checkForRep names
-  infos <- mapM reifyWithWarning names
+  infos <- mapM reifyWithLocals names
   dinfos <- mapM dsInfo infos
   ddecs <- promoteM_ [] $ mapM_ promoteInfo dinfos
   return $ decsToTH ddecs
@@ -64,11 +69,41 @@ promoteOnly qdec = do
   promDecls <- promoteM_ decls $ promoteDecs ddecls
   return $ decsToTH promDecls
 
--- | Generate defunctionalization symbols for existing type family
+-- | Generate defunctionalization symbols for existing type families.
+--
+-- 'genDefunSymbols' has reasonable support for type families that use
+-- dependent quantification. For instance, this:
+--
+-- @
+-- type family MyProxy k (a :: k) :: Type where
+--   MyProxy k (a :: k) = Proxy a
+--
+-- $('genDefunSymbols' [''MyProxy])
+-- @
+--
+-- Will generate the following defunctionalization symbols:
+--
+-- @
+-- data MyProxySym0     :: Type  ~> k ~> Type
+-- data MyProxySym1 (k  :: Type) :: k ~> Type
+-- @
+--
+-- Note that @MyProxySym0@ is a bit more general than it ought to be, since
+-- there is no dependency between the first kind (@Type@) and the second kind
+-- (@k@). But this would require the ability to write something like:
+--
+-- @
+-- data MyProxySym0 :: forall (k :: Type) ~> k ~> Type
+-- @
+--
+-- Which currently isn't possible. So for the time being, the kind of
+-- @MyProxySym0@ will be slightly more general, which means that under rare
+-- circumstances, you may have to provide extra type signatures if you write
+-- code which exploits the dependency in @MyProxy@'s kind.
 genDefunSymbols :: DsMonad q => [Name] -> q [Dec]
 genDefunSymbols names = do
   checkForRep names
-  infos <- mapM (dsInfo <=< reifyWithWarning) names
+  infos <- mapM (dsInfo <=< reifyWithLocals) names
   decs <- promoteMDecs [] $ concatMapM defunInfo infos
   return $ decsToTH decs
 
@@ -106,26 +141,28 @@ promoteShowInstances = concatMapM promoteShowInstance
 
 -- | Produce an instance for 'PShow' from the given type
 promoteShowInstance :: DsMonad q => Name -> q [Dec]
-promoteShowInstance = promoteInstance (mkShowInstance ForPromotion) "Show"
+promoteShowInstance = promoteInstance mkShowInstance "Show"
 
 -- | Produce an instance for @(==)@ (type-level equality) from the given type
 promoteEqInstance :: DsMonad q => Name -> q [Dec]
 promoteEqInstance name = do
   (tvbs, cons) <- getDataD "I cannot make an instance of (==) for it." name
-  cons' <- concatMapM dsCon cons
   tvbs' <- mapM dsTvb tvbs
-  kind <- promoteType (foldType (DConT name) (map tvbToType tvbs'))
+  let data_ty = foldTypeTvbs (DConT name) tvbs'
+  cons' <- concatMapM (dsCon tvbs' data_ty) cons
+  kind <- promoteType (foldTypeTvbs (DConT name) tvbs')
   inst_decs <- mkEqTypeInstance kind cons'
   return $ decsToTH inst_decs
 
-promoteInstance :: DsMonad q => (Maybe DCxt -> DType -> [DCon] -> q UInstDecl)
-                -> String -> Name -> q [Dec]
+promoteInstance :: DsMonad q => DerivDesc q -> String -> Name -> q [Dec]
 promoteInstance mk_inst class_name name = do
   (tvbs, cons) <- getDataD ("I cannot make an instance of " ++ class_name
                             ++ " for it.") name
-  cons' <- concatMapM dsCon cons
   tvbs' <- mapM dsTvb tvbs
-  raw_inst <- mk_inst Nothing (foldType (DConT name) (map tvbToType tvbs')) cons'
+  let data_ty   = foldTypeTvbs (DConT name) tvbs'
+  cons' <- concatMapM (dsCon tvbs' data_ty) cons
+  let data_decl = DataDecl name tvbs' cons'
+  raw_inst <- mk_inst Nothing data_ty data_decl
   decs <- promoteM_ [] $ void $ promoteInstanceDec Map.empty raw_inst
   return $ decsToTH decs
 
@@ -167,7 +204,7 @@ promoteInfo (DPatSynI {}) =
 -- each defining equation of foo uses it the same way. The foo
 -- function will be promoted to a type familty Foo like this:
 --
---   type family Foo (n :: Nat) :: TyFun Bool Bool -> * where
+--   type family Foo (n :: Nat) :: Bool ~> Bool where
 --      Foo Zero = Id
 --      Foo a    = Not
 --
@@ -181,17 +218,21 @@ promoteDecs :: [DDec] -> PrM ()
 promoteDecs raw_decls = do
   decls <- expand raw_decls     -- expand type synonyms
   checkForRepInDecls decls
-  PDecs { pd_let_decs          = let_decs
-        , pd_class_decs        = classes
-        , pd_instance_decs     = insts
-        , pd_data_decs         = datas
-        , pd_derived_eq_decs   = derived_eq_decs } <- partitionDecs decls
+  PDecs { pd_let_decs                = let_decs
+        , pd_class_decs              = classes
+        , pd_instance_decs           = insts
+        , pd_data_decs               = datas
+        , pd_ty_syn_decs             = ty_syns
+        , pd_open_type_family_decs   = o_tyfams
+        , pd_closed_type_family_decs = c_tyfams
+        , pd_derived_eq_decs         = derived_eq_decs } <- partitionDecs decls
 
+  defunTypeDecls ty_syns c_tyfams o_tyfams
     -- promoteLetDecs returns LetBinds, which we don't need at top level
   _ <- promoteLetDecs noPrefix let_decs
   mapM_ promoteClassDec classes
-  let all_meth_sigs = foldMap (lde_types . cd_lde) classes
-  mapM_ (promoteInstanceDec all_meth_sigs) insts
+  let orig_meth_sigs = foldMap (lde_types . cd_lde) classes
+  mapM_ (promoteInstanceDec orig_meth_sigs) insts
   mapM_ promoteDerivedEqDec   derived_eq_decs
   promoteDataDecs datas
 
@@ -202,9 +243,8 @@ promoteDataDecs data_decs = do
   mapM_ promoteDataDec data_decs
   where
     extract_rec_selectors :: DataDecl -> PrM [DLetDec]
-    extract_rec_selectors (DataDecl _nd data_name tvbs cons _derivings) =
-      let arg_ty = foldType (DConT data_name)
-                            (map tvbToType tvbs)
+    extract_rec_selectors (DataDecl data_name tvbs cons) =
+      let arg_ty = foldTypeTvbs (DConT data_name) tvbs
       in
       getRecordSelectors arg_ty cons
 
@@ -237,8 +277,8 @@ promoteLetDecs prefixes decls = do
 --
 --  * for each nullary data constructor we generate a type synonym
 promoteDataDec :: DataDecl -> PrM ()
-promoteDataDec (DataDecl _nd name tvbs ctors _derivings) = do
-  ctorSyms <- buildDefunSymsDataD name tvbs ctors
+promoteDataDec (DataDecl _name _tvbs ctors) = do
+  ctorSyms <- buildDefunSymsDataD ctors
   emitDecs ctorSyms
 
 -- Note [CUSKification]
@@ -248,7 +288,7 @@ promoteDataDec (DataDecl _nd name tvbs ctors _derivings) = do
 -- seems to be no way to avoid this, so we embrace it:
 --
 --   * If a class type variable has no explicit kind, we make no effort to
---     guess it and default to *. This is OK because before TypeInType we were
+--     guess it and default to *. This is OK because before GHC 8.0, we were
 --     limited by KProxy anyway.
 --
 --   * If a class type variable has an explicit kind, it is preserved.
@@ -267,43 +307,51 @@ promoteClassDec decl@(ClassDecl { cd_cxt  = cxt
                                     , lde_infix = infix_decls } }) = do
   let
     -- a workaround for GHC Trac #12928; see Note [CUSKification]
-    cuskify :: DTyVarBndr -> DTyVarBndr
-    cuskify (DPlainTV tvname) = DKindedTV tvname DStarT
-    cuskify tv                = tv
     tvbs = map cuskify tvbs'
   let pClsName = promoteClassName cls_name
   pCxt <- mapM promote_superclass_pred cxt
-  sig_decs <- mapM (uncurry promote_sig) (Map.toList meth_sigs)
-  let defaults_list  = Map.toList defaults
-      defaults_names = map fst defaults_list
-  (default_decs, ann_rhss, prom_rhss)
-    <- mapAndUnzip3M (promoteMethod Nothing meth_sigs) defaults_list
+  forallBind cls_kvs_to_bind $ do
+    sig_decs <- mapM (uncurry promote_sig) (Map.toList meth_sigs)
+    let defaults_list  = Map.toList defaults
+        defaults_names = map fst defaults_list
+    (default_decs, ann_rhss, prom_rhss)
+      <- mapAndUnzip3M (promoteMethod Map.empty Nothing meth_sigs) defaults_list
 
-  let infix_decls' = catMaybes $ map (uncurry promoteInfixDecl) infix_decls
+    let infix_decls' = catMaybes $ map (uncurry promoteInfixDecl)
+                                 $ Map.toList infix_decls
 
-  -- no need to do anything to the fundeps. They work as is!
-  emitDecs [DClassD pCxt pClsName tvbs fundeps
-                    (sig_decs ++ default_decs ++ infix_decls')]
-  let defaults_list' = zip defaults_names ann_rhss
-      proms          = zip defaults_names prom_rhss
-  return (decl { cd_lde = lde { lde_defns = Map.fromList defaults_list'
-                              , lde_proms = Map.fromList proms } })
+    -- no need to do anything to the fundeps. They work as is!
+    emitDecs [DClassD pCxt pClsName tvbs fundeps
+                      (sig_decs ++ default_decs ++ infix_decls')]
+    let defaults_list'   = zip defaults_names ann_rhss
+        proms            = zip defaults_names prom_rhss
+        cls_kvs_to_bind' = cls_kvs_to_bind <$ meth_sigs
+    return (decl { cd_lde = lde { lde_defns     = Map.fromList defaults_list'
+                                , lde_proms     = Map.fromList proms
+                                , lde_bound_kvs = cls_kvs_to_bind' } })
   where
+    cls_kvb_names, cls_tvb_names, cls_kvs_to_bind :: Set Name
+    cls_kvb_names   = foldMap (foldMap fvDType . extractTvbKind) tvbs'
+    cls_tvb_names   = Set.fromList $ map extractTvbName tvbs'
+    cls_kvs_to_bind = cls_kvb_names `Set.union` cls_tvb_names
+
     promote_sig :: Name -> DType -> PrM DDec
     promote_sig name ty = do
       let proName = promoteValNameLhs name
       (argKs, resK) <- promoteUnraveled ty
       args <- mapM (const $ qNewName "arg") argKs
-      emitDecsM $ defunctionalize proName (map Just argKs) (Just resK)
+      let tvbs = zipWith DKindedTV args argKs
+      emitDecsM $ defunReifyFixity proName tvbs (Just resK)
 
       return $ DOpenTypeFamilyD (DTypeFamilyHead proName
-                                                 (zipWith DKindedTV args argKs)
+                                                 tvbs
                                                  (DKindSig resK)
                                                  Nothing)
 
     promote_superclass_pred :: DPred -> PrM DPred
     promote_superclass_pred = go
       where
+      go (DForallPr {}) = fail "Cannot promote quantified constraints"
       go (DAppPr pr ty) = DAppPr <$> go pr <*> promoteType ty
       go (DSigPr pr _k) = go pr    -- just ignore the kind; it can't matter
       go (DVarPr name)  = fail $ "Cannot promote ConstraintKinds variables like "
@@ -313,17 +361,20 @@ promoteClassDec decl@(ClassDecl { cd_cxt  = cxt
 
 -- returns (unpromoted method name, ALetDecRHS) pairs
 promoteInstanceDec :: Map Name DType -> UInstDecl -> PrM AInstDecl
-promoteInstanceDec meth_sigs
+promoteInstanceDec orig_meth_sigs
                    decl@(InstDecl { id_name     = cls_name
                                   , id_arg_tys  = inst_tys
+                                  , id_sigs     = inst_sigs
                                   , id_meths    = meths }) = do
   cls_tvb_names <- lookup_cls_tvb_names
   inst_kis <- mapM promoteType inst_tys
-  let subst = Map.fromList $ zip cls_tvb_names inst_kis
-  (meths', ann_rhss, _) <- mapAndUnzip3M (promoteMethod (Just subst) meth_sigs) meths
-  emitDecs [DInstanceD Nothing [] (foldType (DConT pClsName)
-                                    inst_kis) meths']
-  return (decl { id_meths = zip (map fst meths) ann_rhss })
+  let kvs_to_bind = foldMap fvDType inst_kis
+  forallBind kvs_to_bind $ do
+    let subst = Map.fromList $ zip cls_tvb_names inst_kis
+    (meths', ann_rhss, _) <- mapAndUnzip3M (promoteMethod inst_sigs (Just subst) orig_meth_sigs) meths
+    emitDecs [DInstanceD Nothing [] (foldType (DConT pClsName)
+                                      inst_kis) meths']
+    return (decl { id_meths = zip (map fst meths) ann_rhss })
   where
     pClsName = promoteClassName cls_name
 
@@ -379,27 +430,21 @@ dsReifyTypeNameInfo, which first calls lookupTypeName (to ensure we can find a N
 that's in the type namespace) and _then_ reifies it.
 -}
 
-promoteMethod :: Maybe (Map Name DKind)
+promoteMethod :: Map Name DType -- InstanceSigs for methods
+              -> Maybe (Map Name DKind)
                     -- ^ instantiations for class tyvars (Nothing for default decls)
                     --   See Note [Promoted class method kinds]
               -> Map Name DType     -- method types
               -> (Name, ULetDecRHS)
               -> PrM (DDec, ALetDecRHS, DType)
                  -- returns (type instance, ALetDecRHS, promoted RHS)
-promoteMethod m_subst sigs_map (meth_name, meth_rhs) = do
-  (arg_kis, res_ki) <- lookup_meth_ty
+promoteMethod inst_sigs_map m_subst orig_sigs_map (meth_name, meth_rhs) = do
+  (meth_arg_kis, meth_res_ki) <- lookup_meth_ty
   ((_, _, _, eqns), _defuns, ann_rhs)
-    <- promoteLetDecRHS (Just (arg_kis, res_ki)) sigs_map noPrefix meth_name meth_rhs
-  meth_arg_tvs <- mapM (const $ qNewName "a") arg_kis
-  let -- If we're dealing with an associated type family instance, substitute
-      -- in the kind of the instance for better kind information in the RHS
-      -- helper function. If we're dealing with a default family implementation
-      -- (m_subst = Nothing), there's no need for a substitution.
-      -- See Note [Promoted class method kinds]
-      do_subst      = maybe id substKind m_subst
-      meth_arg_kis' = map do_subst arg_kis
-      meth_res_ki'  = do_subst res_ki
-      helperNameBase = case nameBase proName of
+    <- promoteLetDecRHS (Just (meth_arg_kis, meth_res_ki)) Map.empty Map.empty
+                        noPrefix meth_name meth_rhs
+  meth_arg_tvs <- mapM (const $ qNewName "a") meth_arg_kis
+  let helperNameBase = case nameBase proName of
                          first:_ | not (isHsLetter first) -> "TFHelper"
                          alpha                            -> alpha
 
@@ -417,13 +462,14 @@ promoteMethod m_subst sigs_map (meth_name, meth_rhs) = do
       -- strictly necessary, as kind inference can figure them out just as well.
       family_args = map DVarT meth_arg_tvs
   helperName <- newUniqueName helperNameBase
+  let tvbs = zipWith DKindedTV meth_arg_tvs meth_arg_kis
   emitDecs [DClosedTypeFamilyD (DTypeFamilyHead
                                   helperName
-                                  (zipWith DKindedTV meth_arg_tvs meth_arg_kis')
-                                  (DKindSig meth_res_ki')
+                                  tvbs
+                                  (DKindSig meth_res_ki)
                                   Nothing)
                                eqns]
-  emitDecsM (defunctionalize helperName (map Just meth_arg_kis') (Just meth_res_ki'))
+  emitDecsM (defunctionalize helperName Nothing tvbs (Just meth_res_ki))
   return ( DTySynInstD
              proName
              (DTySynEqn family_args
@@ -434,19 +480,38 @@ promoteMethod m_subst sigs_map (meth_name, meth_rhs) = do
     proName = promoteValNameLhs meth_name
 
     lookup_meth_ty :: PrM ([DKind], DKind)
-    lookup_meth_ty = case Map.lookup meth_name sigs_map of
-      Nothing -> do
-        mb_info <- dsReifyTypeNameInfo proName
-                   -- See Note [Using dsReifyTypeNameInfo when promoting instances]
-        case mb_info of
-          Just (DTyConI (DOpenTypeFamilyD (DTypeFamilyHead _ tvbs mb_res_ki _)) _)
-            -> let arg_kis = map (default_to_star . extractTvbKind) tvbs
-                   res_ki  = default_to_star (resultSigToMaybeKind mb_res_ki)
-                in return (arg_kis, res_ki)
-          _ -> fail $ "Cannot find type annotation for " ++ show proName
-      Just ty -> promoteUnraveled ty
+    lookup_meth_ty =
+      case Map.lookup meth_name inst_sigs_map of
+        Just ty ->
+          -- We have an InstanceSig. These are easy: no substitution for clas
+          -- variables is required at all!
+          promoteUnraveled ty
+        Nothing -> do
+          -- We don't have an InstanceSig, so we must compute the kind to use
+          -- ourselves (possibly substituting for class variables below).
+          (arg_kis, res_ki) <-
+            case Map.lookup meth_name orig_sigs_map of
+              Nothing -> do
+                mb_info <- dsReifyTypeNameInfo proName
+                           -- See Note [Using dsReifyTypeNameInfo when promoting instances]
+                case mb_info of
+                  Just (DTyConI (DOpenTypeFamilyD (DTypeFamilyHead _ tvbs mb_res_ki _)) _)
+                    -> let arg_kis = map (default_to_star . extractTvbKind) tvbs
+                           res_ki  = default_to_star (resultSigToMaybeKind mb_res_ki)
+                        in return (arg_kis, res_ki)
+                  _ -> fail $ "Cannot find type annotation for " ++ show proName
+              Just ty -> promoteUnraveled ty
+          let -- If we're dealing with an associated type family instance, substitute
+              -- in the kind of the instance for better kind information in the RHS
+              -- helper function. If we're dealing with a default family implementation
+              -- (m_subst = Nothing), there's no need for a substitution.
+              -- See Note [Promoted class method kinds]
+              do_subst      = maybe id substKind m_subst
+              meth_arg_kis' = map do_subst arg_kis
+              meth_res_ki'  = do_subst res_ki
+          pure (meth_arg_kis', meth_res_ki')
 
-    default_to_star Nothing  = DStarT
+    default_to_star Nothing  = DConT typeKindName
     default_to_star (Just k) = k
 
 {-
@@ -484,22 +549,25 @@ promoted method implementations like MHelper2.
 promoteLetDecEnv :: (String, String) -> ULetDecEnv -> PrM ([DDec], ALetDecEnv)
 promoteLetDecEnv prefixes (LetDecEnv { lde_defns = value_env
                                      , lde_types = type_env
-                                     , lde_infix = infix_decls }) = do
-  let infix_decls'  = catMaybes $ map (uncurry promoteInfixDecl) infix_decls
+                                     , lde_infix = fix_env }) = do
+  let infix_decls = catMaybes $ map (uncurry promoteInfixDecl)
+                              $ Map.toList fix_env
 
     -- promote all the declarations, producing annotated declarations
   let (names, rhss) = unzip $ Map.toList value_env
   (payloads, defun_decss, ann_rhss)
-    <- fmap unzip3 $ zipWithM (promoteLetDecRHS Nothing type_env prefixes) names rhss
+    <- fmap unzip3 $ zipWithM (promoteLetDecRHS Nothing type_env fix_env prefixes) names rhss
 
   emitDecs $ concat defun_decss
-  let decs = map payload_to_dec payloads ++ infix_decls'
+  bound_kvs <- allBoundKindVars
+  let decs = map payload_to_dec payloads ++ infix_decls
 
     -- build the ALetDecEnv
-  let let_dec_env' = LetDecEnv { lde_defns = Map.fromList $ zip names ann_rhss
-                               , lde_types = type_env
-                               , lde_infix = infix_decls
-                               , lde_proms = Map.empty }  -- filled in promoteLetDecs
+  let let_dec_env' = LetDecEnv { lde_defns     = Map.fromList $ zip names ann_rhss
+                               , lde_types     = type_env
+                               , lde_infix     = fix_env
+                               , lde_proms     = Map.empty   -- filled in promoteLetDecs
+                               , lde_bound_kvs = Map.fromList $ map (, bound_kvs) names }
 
   return (decs, let_dec_env')
   where
@@ -509,14 +577,22 @@ promoteLetDecEnv prefixes (LetDecEnv { lde_defns = value_env
       where
         sig = maybe DNoSig DKindSig m_ki
 
-promoteInfixDecl :: Fixity -> Name -> Maybe DDec
-promoteInfixDecl fixity name
- | isDataConName name || not (isHsLetter (head (nameBase name)))
- = Nothing -- No need to promote fixity declarations for constructor names or
-           -- infix names, as those fixity declarations apply to both
-           -- the value and type namespaces.
+promoteInfixDecl :: Name -> Fixity -> Maybe DDec
+promoteInfixDecl name fixity
+ | nameBase name == nameBase promoted_name
+   -- If a name and its promoted counterpart are the same (modulo module
+   -- prefixes), then there's no need to promote a fixity declaration for
+   -- that name, since the existing fixity declaration will cover both
+   -- the term- and type-level versions of that name,
+   --
+   -- Names that fall into this category include data constructor names
+   -- and infix names, with the exceptions of (.) and (!).
+   -- See Note [Special cases for (.) and (!)] in Data.Singletons.Names.
+ = Nothing
  | otherwise
- = Just $ DLetDec $ DInfixD fixity (promoteValNameLhs name)
+ = Just $ DLetDec $ DInfixD fixity promoted_name
+ where
+  promoted_name = promoteValNameLhs name
 
 -- This function is used both to promote class method defaults and normal
 -- let bindings. Thus, it can't quite do all the work locally and returns
@@ -524,13 +600,14 @@ promoteInfixDecl fixity name
 promoteLetDecRHS :: Maybe ([DKind], DKind)  -- the promoted type of the RHS (if known)
                                             -- needed to fix #136
                  -> Map Name DType       -- local type env't
+                 -> Map Name Fixity      -- local fixity env't
                  -> (String, String)     -- let-binding prefixes
                  -> Name                 -- name of the thing being promoted
                  -> ULetDecRHS           -- body of the thing
                  -> PrM ( (Name, [DTyVarBndr], Maybe DKind, [DTySynEqn]) -- "type family"
                         , [DDec]        -- defunctionalization
                         , ALetDecRHS )  -- annotated RHS
-promoteLetDecRHS m_rhs_ki type_env prefixes name (UValue exp) = do
+promoteLetDecRHS m_rhs_ki type_env fix_env prefixes name (UValue exp) = do
   (res_kind, num_arrows)
     <- case m_rhs_ki of
          Just (arg_kis, res_ki) -> return ( Just (ravelTyFun (arg_kis ++ [res_ki]))
@@ -543,10 +620,13 @@ promoteLetDecRHS m_rhs_ki type_env prefixes name (UValue exp) = do
   case num_arrows of
     0 -> do
       all_locals <- allLocals
-      (exp', ann_exp) <- promoteExp exp
+      let lde_kvs_to_bind = foldMap fvDType res_kind
+      (exp', ann_exp) <- forallBind lde_kvs_to_bind $ promoteExp exp
       let proName = promoteValNameLhsPrefix prefixes name
-      defuns <- defunctionalize proName (map (const Nothing) all_locals) res_kind
-      return ( ( proName, map DPlainTV all_locals, res_kind
+          m_fixity = Map.lookup name fix_env
+          tvbs = map DPlainTV all_locals
+      defuns <- defunctionalize proName m_fixity tvbs res_kind
+      return ( ( proName, tvbs, res_kind
                , [DTySynEqn (map DVarT all_locals) exp'] )
              , defuns
              , AValue (foldType (DConT proName) (map DVarT all_locals))
@@ -555,10 +635,10 @@ promoteLetDecRHS m_rhs_ki type_env prefixes name (UValue exp) = do
       names <- replicateM num_arrows (newUniqueName "a")
       let pats    = map DVarPa names
           newArgs = map DVarE  names
-      promoteLetDecRHS m_rhs_ki type_env prefixes name
+      promoteLetDecRHS m_rhs_ki type_env fix_env prefixes name
                        (UFunction [DClause pats (foldExp exp newArgs)])
 
-promoteLetDecRHS m_rhs_ki type_env prefixes name (UFunction clauses) = do
+promoteLetDecRHS m_rhs_ki type_env fix_env prefixes name (UFunction clauses) = do
   numArgs <- count_args clauses
   (m_argKs, m_resK, ty_num_args) <- case m_rhs_ki of
     Just (arg_kis, res_ki) -> return (map Just arg_kis, Just res_ki, length arg_kis)
@@ -573,28 +653,37 @@ promoteLetDecRHS m_rhs_ki type_env prefixes name (UFunction clauses) = do
 
       |  otherwise
       -> return (replicate numArgs Nothing, Nothing, numArgs)
-  let proName = promoteValNameLhsPrefix prefixes name
+  let proName  = promoteValNameLhsPrefix prefixes name
+      m_fixity = Map.lookup name fix_env
   all_locals <- allLocals
-  defun_decs <- defunctionalize proName
-                (map (const Nothing) all_locals ++ m_argKs) m_resK
   let local_tvbs = map DPlainTV all_locals
   tyvarNames <- mapM (const $ qNewName "a") m_argKs
-  expClauses <- mapM (etaExpand (ty_num_args - numArgs)) clauses
-  (eqns, ann_clauses) <- mapAndUnzipM promoteClause expClauses
-  prom_fun <- lookupVarE name
   let args     = zipWith inferMaybeKindTV tyvarNames m_argKs
       all_args = local_tvbs ++ args
+  defun_decs <- defunctionalize proName m_fixity all_args m_resK
+  expClauses <- mapM (etaContractOrExpand ty_num_args numArgs) clauses
+  let lde_kvs_to_bind = foldMap (foldMap fvDType) m_argKs <> foldMap fvDType m_resK
+  (eqns, ann_clauses) <- forallBind lde_kvs_to_bind $
+                         mapAndUnzipM promoteClause expClauses
+  prom_fun <- lookupVarE name
   return ( (proName, all_args, m_resK, eqns)
          , defun_decs
          , AFunction prom_fun ty_num_args ann_clauses )
 
   where
-    etaExpand :: Int -> DClause -> PrM DClause
-    etaExpand n (DClause pats exp) = do
-      names <- replicateM n (newUniqueName "a")
-      let newPats = map DVarPa names
-          newArgs = map DVarE  names
-      return $ DClause (pats ++ newPats) (foldExp exp newArgs)
+    etaContractOrExpand :: Int -> Int -> DClause -> PrM DClause
+    etaContractOrExpand ty_num_args clause_num_args (DClause pats exp)
+      | n >= 0 = do -- Eta-expand
+          names <- replicateM n (newUniqueName "a")
+          let newPats = map DVarPa names
+              newArgs = map DVarE  names
+          return $ DClause (pats ++ newPats) (foldExp exp newArgs)
+      | otherwise = do -- Eta-contract
+          let (clausePats, lamPats) = splitAt ty_num_args pats
+          lamExp <- mkDLamEFromDPats lamPats exp
+          return $ DClause clausePats lamExp
+      where
+        n = ty_num_args - clause_num_args
 
     count_args (DClause pats _ : _) = return $ length pats
     count_args _ = fail $ "Impossible! A function without clauses."
@@ -603,51 +692,62 @@ promoteClause :: DClause -> PrM (DTySynEqn, ADClause)
 promoteClause (DClause pats exp) = do
   -- promoting the patterns creates variable bindings. These are passed
   -- to the function promoted the RHS
-  (types, new_vars) <- evalForPair $ mapM promotePat pats
-  (ty, ann_exp) <- lambdaBind new_vars $ promoteExp exp
+  ((types, pats'), prom_pat_infos) <- evalForPair $ mapAndUnzipM promotePat pats
+  let PromDPatInfos { prom_dpat_vars    = new_vars
+                    , prom_dpat_sig_kvs = sig_kvs } = prom_pat_infos
+  (ty, ann_exp) <- forallBind sig_kvs $
+                   lambdaBind new_vars $
+                   promoteExp exp
   all_locals <- allLocals   -- these are bound *outside* of this clause
   return ( DTySynEqn (map DVarT all_locals ++ types) ty
-         , ADClause new_vars pats ann_exp )
+         , ADClause new_vars pats' ann_exp )
 
 promoteMatch :: DMatch -> PrM (DTySynEqn, ADMatch)
 promoteMatch (DMatch pat exp) = do
   -- promoting the patterns creates variable bindings. These are passed
   -- to the function promoted the RHS
-  (ty, new_vars) <- evalForPair $ promotePat pat
-  (rhs, ann_exp) <- lambdaBind new_vars $ promoteExp exp
+  ((ty, pat'), prom_pat_infos) <- evalForPair $ promotePat pat
+  let PromDPatInfos { prom_dpat_vars    = new_vars
+                    , prom_dpat_sig_kvs = sig_kvs } = prom_pat_infos
+  (rhs, ann_exp) <- forallBind sig_kvs $
+                    lambdaBind new_vars $
+                    promoteExp exp
   all_locals <- allLocals
   return $ ( DTySynEqn (map DVarT all_locals ++ [ty]) rhs
-           , ADMatch new_vars pat ann_exp)
+           , ADMatch new_vars pat' ann_exp)
 
 -- promotes a term pattern into a type pattern, accumulating bound variable names
-promotePat :: DPat -> QWithAux VarPromotions PrM DType
-promotePat (DLitPa lit) = do
-  lit' <- promoteLitPat lit
-  return lit'
+promotePat :: DPat -> QWithAux PromDPatInfos PrM (DType, ADPat)
+promotePat (DLitPa lit) = (, ADLitPa lit) <$> promoteLitPat lit
 promotePat (DVarPa name) = do
       -- term vars can be symbols... type vars can't!
   tyName <- mkTyName name
-  addElement (name, tyName)
-  return $ DVarT tyName
+  tell $ PromDPatInfos [(name, tyName)] Set.empty
+  return (DVarT tyName, ADVarPa name)
 promotePat (DConPa name pats) = do
-  types <- mapM promotePat pats
+  (types, pats') <- mapAndUnzipM promotePat pats
   let name' = unboxed_tuple_to_tuple name
-  return $ foldType (DConT name') types
+  return (foldType (DConT name') types, ADConPa name pats')
   where
     unboxed_tuple_to_tuple n
       | Just deg <- unboxedTupleNameDegree_maybe n = tupleDataName deg
       | otherwise                                  = n
 promotePat (DTildePa pat) = do
   qReportWarning "Lazy pattern converted into regular pattern in promotion"
-  promotePat pat
+  second ADTildePa <$> promotePat pat
 promotePat (DBangPa pat) = do
   qReportWarning "Strict pattern converted into regular pattern in promotion"
-  promotePat pat
+  second ADBangPa <$> promotePat pat
 promotePat (DSigPa pat ty) = do
-  promoted <- promotePat pat
+  -- We must maintain the invariant that any promoted pattern signature must
+  -- not have any wildcards in the underlying pattern.
+  -- See Note [Singling pattern signatures].
+  wildless_pat <- removeWilds pat
+  (promoted, pat') <- promotePat wildless_pat
   ki <- promoteType ty
-  return $ DSigT promoted ki
-promotePat DWildPa = return DWildCardT
+  tell $ PromDPatInfos [] (fvDType ki)
+  return (DSigT promoted ki, ADSigPa promoted pat' ki)
+promotePat DWildPa = return (DWildCardT, ADWildPa)
 
 promoteExp :: DExp -> PrM (DType, ADExp)
 promoteExp (DVarE name) = fmap (, ADVarE name) $ lookupVarE name
@@ -677,7 +777,7 @@ promoteExp (DLamE names exp) = do
                                  Nothing)
                                [DTySynEqn (map DVarT (all_locals ++ tyNames))
                                           rhs]]
-  emitDecsM $ defunctionalize lambdaName (map (const Nothing) all_args) Nothing
+  emitDecsM $ defunctionalize lambdaName Nothing tvbs Nothing
   let promLambda = foldl apply (DConT (promoteTySym lambdaName 0))
                                (map DVarT all_locals)
   return (promLambda, ADLamE tyNames promLambda names ann_exp)
@@ -704,7 +804,7 @@ promoteExp (DLetE decs exp) = do
 promoteExp (DSigE exp ty) = do
   (exp', ann_exp) <- promoteExp exp
   ty' <- promoteType ty
-  return (DSigT exp' ty', ADSigE ann_exp ty)
+  return (DSigT exp' ty', ADSigE exp' ann_exp ty)
 promoteExp e@(DStaticE _) = fail ("Static expressions cannot be promoted: " ++ show e)
 
 promoteLitExp :: Quasi q => Lit -> q DType
@@ -733,7 +833,8 @@ promoteLitPat lit =
 
 -- See Note [DerivedDecl]
 promoteDerivedEqDec :: DerivedEqDecl -> PrM ()
-promoteDerivedEqDec (DerivedDecl { ded_type = ty, ded_cons = cons }) = do
+promoteDerivedEqDec (DerivedDecl { ded_type = ty
+                                 , ded_decl = DataDecl _ _ cons }) = do
   kind <- promoteType ty
   inst_decs <- mkEqTypeInstance kind cons
   emitDecs inst_decs
