@@ -25,9 +25,13 @@ import Control.Monad.Reader hiding ( mapM )
 import qualified Data.Map as Map
 import Data.List.NonEmpty (NonEmpty(..))
 import Data.Map ( Map )
+import qualified Data.Monoid as Monoid
+import Data.Semigroup as Semigroup
 import Data.Foldable
+import Data.Functor.Identity
 import Data.Traversable
 import Data.Generics
+import Data.Maybe
 import Data.Void
 import Control.Monad.Fail ( MonadFail )
 
@@ -42,17 +46,43 @@ basicTypes = [ ''Maybe
 
 boundedBasicTypes :: [Name]
 boundedBasicTypes =
-            [  ''(,)
+            [ ''(,)
             , ''(,,)
             , ''(,,,)
             , ''(,,,,)
             , ''(,,,,,)
             , ''(,,,,,,)
+            , ''Identity
             ] ++ enumBasicTypes
 
 enumBasicTypes :: [Name]
 enumBasicTypes = [ ''Bool, ''Ordering, ''() ]
 
+semigroupBasicTypes :: [Name]
+semigroupBasicTypes
+  = [ ''Dual
+    , ''All
+    , ''Any
+    , ''Sum
+    , ''Product
+    -- , ''Endo      see https://github.com/goldfirere/singletons/issues/82
+    {- , ''Alt       singletons doesn't support higher kinds :(
+                     see https://github.com/goldfirere/singletons/issues/150
+    -}
+
+    , ''Min
+    , ''Max
+    , ''Semigroup.First
+    , ''Semigroup.Last
+    , ''WrappedMonoid
+    ]
+
+monoidBasicTypes :: [Name]
+monoidBasicTypes
+  = [ ''Monoid.First
+    , ''Monoid.Last
+    ]
+
 -- like reportWarning, but generalized to any Quasi
 qReportWarning :: Quasi q => String -> q ()
 qReportWarning = qReport False
@@ -218,17 +248,6 @@ resultSigToMaybeKind (DKindSig k)                = Just k
 resultSigToMaybeKind (DTyVarSig (DPlainTV _))    = Nothing
 resultSigToMaybeKind (DTyVarSig (DKindedTV _ k)) = Just k
 
--- Get argument types from an arrow type. Removing ForallT is an
--- important preprocessing step required by promoteType.
-unravel :: DType -> ([DTyVarBndr], [DPred], [DType], DType)
-unravel (DForallT tvbs cxt ty) =
-  let (tvbs', cxt', tys, res) = unravel ty in
-  (tvbs ++ tvbs', cxt ++ cxt', tys, res)
-unravel (DAppT (DAppT DArrowT t1) t2) =
-  let (tvbs, cxt, tys, res) = unravel t2 in
-  (tvbs, cxt, t1 : tys, res)
-unravel t = ([], [], [], t)
-
 -- Reconstruct arrow kind from the list of kinds
 ravel :: [DType] -> DType -> DType
 ravel []    res  = res
@@ -236,11 +255,12 @@ ravel (h:t) res = DAppT (DAppT DArrowT h) (ravel t res)
 
 -- | Convert a 'DPred' to a 'DType'.
 predToType :: DPred -> DType
-predToType (DAppPr p t) = DAppT (predToType p) t
-predToType (DSigPr p k) = DSigT (predToType p) k
-predToType (DVarPr n)   = DVarT n
-predToType (DConPr n)   = DConT n
-predToType DWildCardPr  = DWildCardT
+predToType (DForallPr tvbs cxt p) = DForallT tvbs cxt (predToType p)
+predToType (DAppPr p t)           = DAppT (predToType p) t
+predToType (DSigPr p k)           = DSigT (predToType p) k
+predToType (DVarPr n)             = DVarT n
+predToType (DConPr n)             = DConT n
+predToType DWildCardPr            = DWildCardT
 
 -- count the number of arguments in a type
 countArgs :: DType -> Int
@@ -280,10 +300,6 @@ substType subst (DForallT tvbs cxt inner_ty)
     (subst', tvbs') = mapAccumL subst_tvb subst tvbs
     cxt'            = map (substPred subst') cxt
     inner_ty'       = substType subst' inner_ty
-
-    subst_tvb s tvb@(DPlainTV n) = (Map.delete n s, tvb)
-    subst_tvb s (DKindedTV n k)  = (Map.delete n s, DKindedTV n (substKind s k))
-
 substType subst (DAppT ty1 ty2) = substType subst ty1 `DAppT` substType subst ty2
 substType subst (DSigT ty ki) = substType subst ty `DSigT` substType subst ki
 substType subst (DVarT n) =
@@ -294,41 +310,47 @@ substType _ ty@(DConT {}) = ty
 substType _ ty@(DArrowT)  = ty
 substType _ ty@(DLitT {}) = ty
 substType _ ty@DWildCardT = ty
-substType _ ty@DStarT     = ty
 
 substPred :: Map Name DType -> DPred -> DPred
 substPred subst pred | Map.null subst = pred
+substPred subst (DForallPr tvbs cxt inner_pred)
+  = DForallPr tvbs' cxt' inner_pred'
+  where
+    (subst', tvbs') = mapAccumL subst_tvb subst tvbs
+    cxt'            = map (substPred subst') cxt
+    inner_pred'     = substPred subst' inner_pred
 substPred subst (DAppPr pred ty) =
   DAppPr (substPred subst pred) (substType subst ty)
-substPred subst (DSigPr pred ki) = DSigPr (substPred subst pred) ki
+substPred subst (DSigPr pred ki) =
+  DSigPr (substPred subst pred) (substKind subst ki)
 substPred _ pred@(DVarPr {}) = pred
 substPred _ pred@(DConPr {}) = pred
 substPred _ pred@DWildCardPr = pred
 
-substKindInPred :: Map Name DKind -> DPred -> DPred
-substKindInPred subst pred | Map.null subst = pred
-substKindInPred subst (DAppPr pred ty) =
-  DAppPr (substKindInPred subst pred) (substType subst ty)
-substKindInPred subst (DSigPr pred ki) = DSigPr (substKindInPred subst pred)
-                                                (substKind subst ki)
-substKindInPred _ pred@(DVarPr {}) = pred
-substKindInPred _ pred@(DConPr {}) = pred
-substKindInPred _ pred@DWildCardPr = pred
+subst_tvb :: Map Name DKind -> DTyVarBndr -> (Map Name DKind, DTyVarBndr)
+subst_tvb s tvb@(DPlainTV n) = (Map.delete n s, tvb)
+subst_tvb s (DKindedTV n k)  = (Map.delete n s, DKindedTV n (substKind s k))
 
-substKindInTvb :: Map Name DKind -> DTyVarBndr -> DTyVarBndr
-substKindInTvb _ tvb@(DPlainTV _) = tvb
-substKindInTvb subst (DKindedTV n ki) = DKindedTV n (substKind subst ki)
-
-addStar :: DKind -> DKind
-addStar t = DAppT (DAppT DArrowT t) DStarT
-
-addStar_maybe :: Maybe DKind -> Maybe DKind
-addStar_maybe = fmap addStar
+cuskify :: DTyVarBndr -> DTyVarBndr
+cuskify (DPlainTV tvname) = DKindedTV tvname $ DConT typeKindName
+cuskify tvb               = tvb
 
 -- apply a type to a list of types
 foldType :: DType -> [DType] -> DType
 foldType = foldl DAppT
 
+-- apply a type to a list of type variable binders
+foldTypeTvbs :: DType -> [DTyVarBndr] -> DType
+foldTypeTvbs ty = foldType ty . map tvbToType
+
+-- apply a pred to a list of types
+foldPred :: DPred -> [DType] -> DPred
+foldPred = foldl DAppPr
+
+-- apply a pred to a list of type variable binders
+foldPredTvbs :: DPred -> [DTyVarBndr] -> DPred
+foldPredTvbs pr = foldPred pr . map tvbToType
+
 -- | Decompose an applied type into its individual components. For example, this:
 --
 -- @
@@ -349,6 +371,13 @@ unfoldType = go []
     go acc (DForallT _ _ t) = go acc t
     go acc t                = t :| acc
 
+-- Construct a data type's variable binders, possibly using fresh variables
+-- from the data type's kind signature.
+buildDataDTvbs :: DsMonad q => [DTyVarBndr] -> Maybe DKind -> q [DTyVarBndr]
+buildDataDTvbs tvbs mk = do
+  extra_tvbs <- mkExtraDKindBinders $ fromMaybe (DConT typeKindName) mk
+  pure $ tvbs ++ extra_tvbs
+
 -- apply an expression to a list of expressions
 foldExp :: DExp -> [DExp] -> DExp
 foldExp = foldl DAppE
@@ -404,7 +433,8 @@ instance (Quasi q, Monoid m) => Quasi (QWithAux m q) where
   qReifyConStrictness = lift `comp1` qReifyConStrictness
   qIsExtEnabled       = lift `comp1` qIsExtEnabled
   qExtsEnabled        = lift qExtsEnabled
-  qAddForeignFile     = lift `comp2` qAddForeignFile
+  qAddForeignFilePath = lift `comp2` qAddForeignFilePath
+  qAddTempFile        = lift `comp1` qAddTempFile
   qAddCorePlugin      = lift `comp1` qAddCorePlugin
 
   qRecover exp handler = do
