@@ -1,14 +1,12 @@
-{-# OPTIONS_GHC -Wno-unused-imports #-}
-
 module Promote.GenDefunSymbols where
 
-import Data.Singletons (Apply, TyFun)
-import Data.Singletons.Promote
+import Data.Singletons (Apply, type (~>))
 import Data.Singletons.SuppressUnusedWarnings
+import Data.Singletons.TH (genDefunSymbols)
 import GHC.TypeLits hiding (type (*))
-import Data.Kind
+import Data.Kind (Type)
 
-type family LiftMaybe (f :: TyFun a b -> *) (x :: Maybe a) :: Maybe b where
+type family LiftMaybe (f :: a ~> b) (x :: Maybe a) :: Maybe b where
     LiftMaybe f Nothing = Nothing
     LiftMaybe f (Just a) = Just (Apply f a)
 
