@@ -1,6 +1,6 @@
 module Singletons.PolyKindsApp where
 
-import Data.Kind
+import Data.Kind (Type)
 import Data.Singletons.TH
 
 $(singletons [d|
